import bpy

from . import draw_voxel
from . import color_voxel
from . import erase_voxel
from . import preview_voxel
from . import tool_helper
from . import utils
from . import vox_preview_voxel_manager
from . import vox_object_settings
from . import vox_gn_setup
from . import vox_manager

from bpy.app.handlers import persistent

bl_info = {
    "name":        "Vox - Voxel Toolset",
    "author":      "Michael Soluyanov (multlabs.com)",
    "version":     (1, 1),
    "blender":     (3, 2, 0),
    "location":    "Edit mode toolbar, object properties",
    "description": "desc",
    "warning":     "",
    "wiki_url":    "---",
    "tracker_url": "???",
    "category":    "Mesh",
}


classes = (
    vox_object_settings.vox_slots,
    vox_object_settings.VOX_UL_slots,
    vox_object_settings.VOX_OP_add_vox_slot,
    vox_object_settings.VOX_OP_remove_vox_slot,
    vox_object_settings.VOX_PT_slot_selector,
    vox_object_settings.vox_object_settings,
    vox_object_settings.VOX_OP_rebuild_model,

    color_voxel.VIEW3D_PT_vox_color_voxel_settings,
    draw_voxel.draw_voxel_op,
    color_voxel.color_voxel_op,
    preview_voxel.preview_voxel_op,
    tool_helper.vox_tool_status_listener,
    utils.get_voxel_color_op,
    vox_object_settings.VOX_PT_object_settings,
    vox_object_settings.VOX_PT_object_settings_panel,
    vox_gn_setup.vox_OP_rebuild_GN_tree
)


@persistent
def load_handler(dummy):
    vox_manager.clearcashe()


def register():
    """
    Method called by Blender when enabling the add-on
    :return: nothing
    """

    for cls in classes:
        bpy.utils.register_class(cls)
    bpy.types.Object.vox_settings = bpy.props.PointerProperty(
        type=vox_object_settings.vox_object_settings
    )

    bpy.utils.register_tool(
        draw_voxel.draw_voxel_tool,
        after={"builtin.measure"},
        separator=True, group=True
    )
    bpy.utils.register_tool(
        color_voxel.color_voxel_tool,
        after={draw_voxel.draw_voxel_tool.bl_idname},
        separator=False, group=True
    )
    bpy.utils.register_tool(
        erase_voxel.erase_voxel_tool,
        after={color_voxel.color_voxel_tool.bl_idname},
        separator=False, group=True
    )
    bpy.app.handlers.depsgraph_update_post.append(load_handler)


def unregister():
    """
    Method called by Blender when disabling or removing the add-on
    :return: nothing
    """
    vox_preview_voxel_manager.resetHelper()

    bpy.utils.unregister_tool(draw_voxel.draw_voxel_tool)
    bpy.utils.unregister_tool(color_voxel.color_voxel_tool)
    bpy.utils.unregister_tool(erase_voxel.erase_voxel_tool)

    for cls in reversed(classes):
        bpy.utils.unregister_class(cls)

    bpy.types.Object.vox_settings = None
    bpy.app.handlers.depsgraph_update_post.remove(load_handler)
