from bpy.types import Operator
import bpy
from . import vox_manager
from . import vox_preview_voxel_manager
from mathutils import Vector
from bl_ui.space_toolsystem_common import ToolSelectPanelHelper


class preview_voxel_op(Operator):
    """Preview voxel position"""
    bl_idname = "mesh.vox_prview_voxel"
    bl_label = "Preview voxel position"

    add_new: bpy.props.BoolProperty(default=True)
    tool_color: bpy.props.FloatVectorProperty(name="Color", subtype="COLOR",
                                              default=[1.0, 1.0, 1.0])

    @classmethod
    def poll(cls, context):
        return (context.mode == 'EDIT_MESH')

    def invoke(self, context, event):
        
        tool = ToolSelectPanelHelper.tool_active_from_context(context)
        if tool.idname == "mesh.vox_draw_voxel_tool":
            props = tool.operator_properties('mesh.vox_draw_voxel')

        elif tool.idname == "mesh.vox_color_voxel_tool":
            props = tool.operator_properties('mesh.vox_color_voxel')
            if props.operatormode == "BRUSH":
                vox_preview_voxel_manager.drawCircle(
                    (event.mouse_region_x, event.mouse_region_y), context, props.radius)
                return {'PASS_THROUGH'}

        location, vect = vox_manager.get_addvox_position(
            context, event.mouse_region_x,
            event.mouse_region_y,
            self.add_new
        )

        global last_pos
        global last_vect

        if not location:
            last_pos = location
            vox_preview_voxel_manager.resetHelper()
            context.area.tag_redraw()
            return {'PASS_THROUGH'}

        if location == last_pos and last_vect == vect:
            return {'PASS_THROUGH'}

        vox_preview_voxel_manager.resetHelper()

        last_pos = location
        last_vect = vect

        shape = 'CUBE'
        if self.add_new:
            shape = 'PLANE' if vect else 'CUBE'
        else:
            shape = 'CUBE' if vect else 'NONE'
        if shape != 'NONE':

            vox_preview_voxel_manager.setHelper(
                location, context, shape, vect, self.tool_color[:])

        context.area.tag_redraw()

        return {'PASS_THROUGH'}
        return {'FINISHED'}


global last_pos
last_pos = Vector((10000, 0, 0))

global last_vect
last_vect = False
