import bmesh
from mathutils import Vector



def add_vertex_volume_nearest(context, start, end, color):
    """
    Adds vertex volume at spectific coordinates
    blender internal functions is used
    :return: created vertex
    """

    obj = context.object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)
    validate_attributes(bm)
    depsgraph = context.view_layer.depsgraph

    x1, y1, z1 = round(min(start[0], end[0])), round(min(
        start[1], end[1])),  round(min(start[2], end[2]))
    x2, y2, z2 = round(max(start[0], end[0])), round(max(
        start[1], end[1])),  round(max(start[2], end[2]))

    if x1 == x2:
        for y in range(y1, y2+1):
            for z in range(z1, z2+1):
                loc, normal, index = obj.closest_point_on_mesh(
                    Vector((x1, y, z)), depsgraph=depsgraph)
                if square_distance(loc, Vector((x1, y, z))) <= 0.5:
                    continue
                add_vertex_bm(bm, Vector((x1, y, z)), color)
    elif y1 == y2:
        for x in range(x1, x2+1):
            for z in range(z1, z2+1):
                loc, normal, index = obj.closest_point_on_mesh(
                    Vector((x, y1, z)), depsgraph=depsgraph)
                if square_distance(loc, Vector((x, y1, z))) <= 0.5:
                    continue
                add_vertex_bm(bm, Vector((x, y1, z)), color)
    elif z1 == z2:
        for x in range(x1, x2+1):
            for y in range(y1, y2+1):
                loc, normal, index = obj.closest_point_on_mesh(
                    Vector((x, y, z1)), depsgraph=depsgraph)
                if square_distance(loc, Vector((x, y, z1))) <= 0.5:
                    continue
                add_vertex_bm(bm, Vector((x, y, z1)), color)

    bmesh.update_edit_mesh(obj.data)


def add_vertex_volume_cahsed(context, start, end, color):
    """
    Adds vertex volume at spectific coordinates
    dictionary cahse is used
    :return: created vertex
    """
    obj = context.object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)
    validate_attributes(bm)
    vertsarray = cashe_vertexes(bm)

    x1, y1, z1 = round(min(start[0], end[0])), round(min(
        start[1], end[1])),  round(min(start[2], end[2]))
    x2, y2, z2 = round(max(start[0], end[0])), round(max(
        start[1], end[1])),  round(max(start[2], end[2]))

    if x1 == x2:
        for y in range(y1, y2+1):
            for z in range(z1, z2+1):
                key = str(x1)+'-'+str(y)+'-'+str(z)
                if key in vertsarray:
                    continue
                vertsarray[key] = add_vertex_bm(
                    bm, Vector((x1, y, z)), color)
    elif y1 == y2:
        for x in range(x1, x2+1):
            for z in range(z1, z2+1):
                key = str(x)+'-'+str(y1)+'-'+str(z)
                if key in vertsarray:
                    continue
                vertsarray[key] = add_vertex_bm(
                    bm, Vector((x, y1, z)), color)
    elif z1 == z2:
        for x in range(x1, x2+1):
            for y in range(y1, y2+1):
                key = str(x)+'-'+str(y)+'-'+str(z1)
                if key in vertsarray:
                    continue
                vertsarray[key] = add_vertex_bm(
                    bm, Vector((x, y, z1)), color)

    bmesh.update_edit_mesh(obj.data)
