from bpy import props
from bpy.types import Operator
from . import vox_manager
from mathutils import Vector
from bl_ui.space_toolsystem_common import ToolSelectPanelHelper


class get_voxel_color_op(Operator):
    """Get original voxel color"""
    bl_idname = "mesh.vox_get_voxel_color"
    bl_label = "Preview voxel position"

    perw_color: Vector((0, 0, 0))
    props

    @classmethod
    def poll(cls, context):
        return (context.mode == 'EDIT_MESH')

    def invoke(self, context, event):

        location, vect = vox_manager.get_addvox_position(
            context, event.mouse_region_x, event.mouse_region_y, False)

        tool = ToolSelectPanelHelper.tool_active_from_context(context)
        if tool.idname == "mesh.vox_draw_voxel_tool":
            self.props = tool.operator_properties('mesh.vox_draw_voxel')
        elif tool.idname == "mesh.vox_color_voxel_tool":
            self.props = tool.operator_properties('mesh.vox_color_voxel')
        else:
            return {'CANCELLED'}

        vox_manager.get_color_prep(context)
        color = vox_manager.get_color(context, location)
        self.perw_color = self.props.color
        self.props.color = color[:3]

        context.area.tag_redraw()

        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

    def modal(self, context, event):
        context.area.tag_redraw()

        if event.type == 'MOUSEMOVE':
            location, vect = vox_manager.get_addvox_position(
                context, event.mouse_region_x, event.mouse_region_y, False)
            color = vox_manager.get_color(context, location)
            self.props.color = color[:3]
            return {'RUNNING_MODAL'}

        elif event.type == 'LEFTMOUSE':

            return {'FINISHED'}

        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            self.props.color = self.perw_color[:3]
            return {'CANCELLED'}

        return {'RUNNING_MODAL'}
