from audioop import reverse
from pickle import FALSE
import bpy
import bmesh


def validate_attributes(bm):
    if 'Vox Color' not in bm.verts.layers.float_color.keys():
        bm.verts.layers.float_color.new('Vox Color')
    if 'Vox Model' not in bm.verts.layers.int.keys():
        bm.verts.layers.int.new('Vox Model')
    if 'Vox Index' not in bm.verts.layers.int.keys():
        bm.verts.layers.int.new('Vox Index')
    if 'Vox Rotation' not in bm.verts.layers.float_vector.keys():
        bm.verts.layers.float_vector.new('Vox Rotation')
    if 'Col' not in bm.loops.layers.color.keys():
        bm.loops.layers.color.new('Col')


def getNode(nodes, name, type):
    if name not in nodes:
        node = nodes.new(type)
        node.name = name
        return node
    else:
        return nodes[name]


def rebuild_def_material(context):
    defmatname = "Vox Default Material"

    if defmatname in bpy.data.materials:
        material = bpy.data.materials[defmatname]
    else:
        material = bpy.data.materials.new(defmatname)
        material.use_nodes = True
        nodes = material.node_tree.nodes
        nodes.clear()

    material.use_nodes = True
    nodes = material.node_tree.nodes
    links = material.node_tree.links
    out = getNode(nodes, "Vox:Material Output", "ShaderNodeOutputMaterial")
    prn = getNode(nodes, "Vox:Principled BSDF", "ShaderNodeBsdfPrincipled")
    atr = getNode(nodes, "Vox:Attribute", "ShaderNodeAttribute")

    out.location = (300, 0)
    atr.location = (-300, 0)

    links.new(atr.outputs[0], prn.inputs[0])
    links.new(prn.outputs[0], out.inputs[0])

    atr.attribute_name = "Col"

    return material


def hide_sockets(node):
    for s in node.inputs:
        if len(s.links) == 0:
            s.hide = True
    for s in node.outputs:
        if len(s.links) == 0:
            s.hide = True


def rebuild_GN_tree(context):
    print('rebuild tree')
    obj = context.object
    origmode = obj.mode
    bpy.ops.object.mode_set(mode='EDIT')

    me = obj.data
    bm = bmesh.from_edit_mesh(me)
    validate_attributes(bm)

    vs = obj.vox_settings
    vs.active = False

    if len(vs.vox_slots) == 0:
        newslot = vs.vox_slots.add()
        newslot.name = 'Slot'+str(len(vs.vox_slots))

    bpy.ops.object.mode_set(mode=origmode)
    gnmod = None
    for gnmod in obj.modifiers:
        if gnmod.type == "NODES":
            break

    if (gnmod is None) or (gnmod.type != "NODES"):
        gnmod = obj.modifiers.new("Vox", "NODES")
        # 3.2
        if gnmod.node_group is None:
            ng = bpy.data.node_groups.new("Vox", "GeometryNodeTree")
            print(ng)
            gnmod.node_group = ng
    obj.modifiers.active = gnmod

    if gnmod.node_group.users > 1:
        gnmod.node_group = gnmod.node_group.copy()

    # outputs:

    outs = gnmod.node_group.outputs
    if "Geometry" not in outs:
        outs.new("NodeSocketGeometry", "Geometry")

    if "Color" not in outs:
        outs.new("NodeSocketColor", "Color")
    id = outs["Color"].identifier
    gnmod[id+'_attribute_name'] = "Col"

    # inputs:
    ins = gnmod.node_group.inputs

    if "Geometry" not in ins:
        ins.new("NodeSocketGeometry", "Geometry")

    if "Color" not in ins:
        ins.new("NodeSocketColor", "Color")
    id = ins["Color"].identifier
    gnmod[id+'_attribute_name'] = "Vox Color"
    gnmod[id+'_use_attribute'] = True

    if "Model" not in ins:
        ins.new("NodeSocketInt", "Model")
    id = ins["Model"].identifier
    gnmod[id+'_attribute_name'] = "Vox Model"
    gnmod[id+'_use_attribute'] = True

    if "Index" not in ins:
        ins.new("NodeSocketInt", "Index")
    id = ins["Index"].identifier
    gnmod[id+'_attribute_name'] = "Vox Index"
    gnmod[id+'_use_attribute'] = True

    if "Rotation" not in ins:
        ins.new("NodeSocketVector", "Rotation")
    id = ins["Rotation"].identifier
    gnmod[id+'_attribute_name'] = "Vox Rotation"
    gnmod[id+'_use_attribute'] = True

    if "Simple Boxes" not in ins:
        ins.new("NodeSocketBool", "Simple Boxes")

    if "Simple Instances" not in ins:
        ins.new("NodeSocketBool", "Simple Instances")

    nodes = gnmod.node_group.nodes
    links = gnmod.node_group.links

    outputn = getNode(nodes, "Group Output", "NodeGroupOutput")
    inputn = getNode(nodes, "Group Input", "NodeGroupInput")
    inputn.location = (-1000, 100)

    # now building the node tree:
    ca = getNode(nodes, "Vox:ColorConvert", "GeometryNodeCaptureAttribute")
    ca.data_type = 'FLOAT_COLOR'
    ca.location = (-800, 100)
    links.new(inputn.outputs["Color"], ca.inputs[3])
    links.new(inputn.outputs["Geometry"], ca.inputs["Geometry"])
    links.new(ca.outputs[3], outputn.inputs["Color"])

    mp = getNode(nodes, "Vox:ToPoints", "GeometryNodeMeshToPoints")
    mp.location = (-600, 300)
    mp.mode = 'VERTICES'
    links.new(ca.outputs[0], mp.inputs[0])

    cube = getNode(nodes, "Vox:defCube", "GeometryNodeMeshCube")
    cube.location = (-400, -100)
    hide_sockets(cube)

    cubem = getNode(nodes, "Vox:defCubeMat", "GeometryNodeSetMaterial")
    cubem.location = (-200, -100)
    links.new(cube.outputs[0], cubem.inputs[0])
    mat = rebuild_def_material(context)
    cubem.inputs[2].default_value = mat

    cubei = getNode(nodes, "Vox:SimpleInstance",
                    "GeometryNodeInstanceOnPoints")
    links.new(cubem.outputs[0], cubei.inputs[2])
    links.new(mp.outputs[0], cubei.inputs[0])
    cubei.location = (0, -100)

    simpr = getNode(nodes, "Vox:SimpleRealize", "GeometryNodeRealizeInstances")
    simpr.location = (200, -100)
    links.new(cubei.outputs[0], simpr.inputs[0])

    # clear slots in case if there was different settings
    for i in range(0, len(obj.vox_settings.vox_slots)+2):
        if "Vox:slotS"+str(i) in nodes:
            nodes.remove(nodes["Vox:slotS"+str(i)])
        if "Vox:slotM"+str(i) in nodes:
            nodes.remove(nodes["Vox:slotM"+str(i)])
        if "Vox:slotI"+str(i) in nodes:
            nodes.remove(nodes["Vox:slotI"+str(i)])
        if "Vox:slotC"+str(i) in nodes:
            nodes.remove(nodes["Vox:slotC"+str(i)])
        if "Vox:slotG"+str(i) in nodes:
            nodes.remove(nodes["Vox:slotG"+str(i)])
    i = 0

    join = getNode(nodes, "Vox:join", "GeometryNodeJoinGeometry")

    for slot in obj.vox_settings.vox_slots:
        xoff = i*400

        idef = getNode(nodes, "Vox:slotI"+str(i), "GeometryNodeInstanceOnPoints")

        if slot.vox_collection is None:
            slcube = getNode(nodes, "Vox:slotS"+str(i), "GeometryNodeMeshCube")
            slcube.location = (xoff-400, 300)
            hide_sockets(slcube)

            slmodel = getNode(nodes, "Vox:slotM"+str(i), "GeometryNodeSetMaterial")
            slmodel.location = (xoff-400, 230)
            links.new(slcube.outputs[0], slmodel.inputs[0])
            mat = rebuild_def_material(context)
            slmodel.inputs[2].default_value = mat
            idef.inputs[3].default_value = False

        else:
            slmodel = getNode(nodes, "Vox:slotS"+str(i), "GeometryNodeCollectionInfo")
            slmodel.location = (xoff-400, 300)
            slmodel.inputs[0].default_value = slot.vox_collection
            slmodel.inputs[1].default_value = True
            slmodel.inputs[2].default_value = True
            idef.inputs[3].default_value = True

        links.new(slmodel.outputs[0], idef.inputs[2])
        links.new(mp.outputs[0], idef.inputs[0])
        links.new(idef.outputs[0], join.inputs[0])
        idef.location = (xoff-200, 400)

        cdef = getNode(nodes, "Vox:slotC"+str(i), "ShaderNodeMath")
        cdef.operation = "COMPARE"
        links.new(cdef.outputs[0], idef.inputs[1])
        cdef.inputs[1].default_value = i
        cdef.location = (xoff-400, 500)

        gi = getNode(nodes, "Vox:slotG"+str(i), "NodeGroupInput")
        links.new(gi.outputs["Model"], cdef.inputs[0])
        links.new(gi.outputs["Rotation"], idef.inputs[5])
        links.new(gi.outputs["Index"], idef.inputs[4])
        hide_sockets(gi)
        gi.location = (xoff-600, 550)

        i += 1

    bm = getNode(nodes, "Vox:and", "FunctionNodeBooleanMath")
    bm.location = (xoff+200, 500)

    sw1 = getNode(nodes, "Vox:switch1", "GeometryNodeSwitch")
    sw1.location = (xoff+200, 300)

    iv = getNode(nodes, "Vox:isView", "GeometryNodeIsViewport")
    links.new(iv.outputs[0], bm.inputs[0])
    iv.location = (xoff+0, 600)

    gi = getNode(nodes, "Vox:groupEnd", "NodeGroupInput")
    links.new(gi.outputs["Simple Boxes"], bm.inputs[1])
    links.new(gi.outputs["Simple Instances"], sw1.inputs[1])

    hide_sockets(gi)
    gi.location = (xoff-0, 500)

    join.location = (xoff-0, 350)

    ri = getNode(nodes, "Vox:endRealizeInstances", "GeometryNodeRealizeInstances")
    ri.location = (xoff-0, 250)
    sw1.input_type = "GEOMETRY"
    links.new(ri.outputs[0], sw1.inputs[14])
    links.new(join.outputs[0], ri.inputs[0])
    links.new(join.outputs[0], sw1.inputs[15])

    sw2 = getNode(nodes, "Vox:switch2", "GeometryNodeSwitch")
    sw2.location = (xoff+400, 300)
    sw2.input_type = "GEOMETRY"
    links.new(bm.outputs[0], sw2.inputs[1])
    links.new(sw1.outputs[6], sw2.inputs[14])
    links.new(simpr.outputs[0], sw2.inputs[15])
    links.new(sw2.outputs[6], outputn.inputs[0])

    hide_sockets(inputn)
    outputn.location = (xoff+600, 100)


class vox_OP_rebuild_GN_tree(bpy.types.Operator):
    """Rebuild Geometry nodes tree"""
    bl_idname = "object.vox_rebuild_gn_tree"
    bl_label = "Rebuild nodes"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return context.active_object is not None

    def execute(self, context):
        rebuild_GN_tree(context)
        return {'FINISHED'}
