import math
from random import randrange
import bmesh
from bpy_extras.view3d_utils import (
    region_2d_to_vector_3d,
    region_2d_to_location_3d,
    location_3d_to_region_2d
)
from mathutils.kdtree import KDTree
from mathutils.bvhtree import BVHTree
from mathutils import Vector
import copy
import bpy
import re
from collections import defaultdict
import string

from . import vox_gn_setup


def vox_offset_to_id(x, y, z):
    return (z+1) * 9 + (y+1) * 3 + (x+1)


def update_vertex_model(bm, kd, vert, models, zero_level_ground):
    """Update used voxel for a specific vert in bmesh"""
    # vert = bm[index]
    nearest_tollerance = 0.1
    vm = bm.verts.layers.int['Vox Model']
    vi = bm.verts.layers.int['Vox Index']
    vr = bm.verts.layers.float_vector['Vox Rotation']

    occupied = ["1"] * 27

    for (co, index, dist) in kd.find_range(vert.co, 1.45):
        off = co - vert.co
        if 0.5-abs(off[0] % 1 - 0.5) > nearest_tollerance:
            continue
        if 0.5-abs(off[1] % 1 - 0.5) > nearest_tollerance:
            continue
        if 0.5-abs(off[2] % 1 - 0.5) > nearest_tollerance:
            continue
        id = vox_offset_to_id(round(off[0]), round(off[1]), round(off[2]))
        occupied[id] = "2" if vert[vm] == bm.verts[index][vm] else "3"

    for z in range(-1, 2):
        if(round(vert.co[2]+z) == -1):

            for x in range(-1, 2):
                for y in range(-1, 2):
                    id = vox_offset_to_id(round(x), round(y), round(z))
                    # occupied[id] = "3"

    occupied = "".join(occupied)

    good_blocks = []
    rangesum = 0
    weight = -1

    for m in models:

        if m['slot'] != vert[vm]:
            continue
        if m['ground']:
            if round(vert.co[2]) != 0:
                continue
        if m['even_x']:
            if round(vert.co[0] % 2) != 0:
                continue
        if m['even_y']:
            if round(vert.co[1] % 2) != 0:
                continue
        if m['even_z']:
            if round(vert.co[2] % 2) != 0:
                continue
        if re.match(m['occupied'], occupied):
            if weight < m['weight']:
                # blocks with more rules more important
                good_blocks = []
                rangesum = 0
            elif weight > m['weight']:
                continue
            weight = m['weight']
            good_blocks.append(m)
            rangesum += m['random']

    if rangesum > 0:
        selected = randrange(0, rangesum)
        rangesum = 0
        for m in good_blocks:
            rangesum += m['random']
            if selected < rangesum:

                vert[vi] = m['block']
                vert[vr] = m['rotation']
                break


def get_model_weigit(block):
    if block.vox_settings.use_custom_weight:
        return block.vox_settings.custom_weight_value
    weight = 0
    for i in range(0, 27):
        if i == 13:
            o = "SAME"
        else:
            o = getattr(block.vox_settings, "occupied"+str(i))
        if (o == "FREE"):
            weight += 101
        if (o == "SAME"):
            weight += 101
        if (o == "ANY"):
            weight += 100
        if (o == "EMPTY"):
            weight += 100

    if block.vox_settings.only_on_ground:
        weight += 250

    if block.vox_settings.even_z:
        weight += 30
    if block.vox_settings.even_z:
        weight += 30
    if block.vox_settings.even_z:
        weight += 30

    return weight-101


def getmodels(context):
    """Cahsing all models into one sorted array"""
    models = []
    obj = context.object

    items = obj.vox_settings.bl_rna.properties['occupied0'].enum_items

    for slotid, slot in enumerate(obj.vox_settings.vox_slots):
        col = slot.vox_collection
        if col is None:
            continue
        blocks = col.all_objects[:]
        for block in blocks:
            block.name = block.name.title()

        blocks.sort(key=lambda x: re.sub('[a-z]', lambda m: m[0].upper(), x.name))

        for blockid, block in enumerate(blocks):

            if not block.vox_settings.active:
                continue
            weight = get_model_weigit(block)
            occupied = ["."] * 27

            for i in range(0, 27):
                if i == 13:
                    o = "SAME"
                else:
                    o = getattr(block.vox_settings, "occupied"+str(i))
                if items[o].value > 0:
                    occupied[i] = str(items[o].value)
                    if (o == "SAME"):
                        pass
                    if (o == "ANY"):
                        occupied[i] = '[23]'
                    if (o == "EMPTY"):
                        occupied[i] = '[13]'
                    if (o == "FREE"):
                        occupied[i] = '1'
            models.append({
                "slot": slotid,
                "block": blockid,
                "occupied": "".join(occupied),
                "rotation": (0, 0, 0),
                "random": block.vox_settings.random_count,
                "weight": weight,
                "ground": block.vox_settings.only_on_ground,
                "even_x": block.vox_settings.even_x,
                "even_y": block.vox_settings.even_y,
                "even_z": block.vox_settings.even_z,
            })

            occupied1 = ["."] * 27
            if block.vox_settings.allow_rotation:
                for a in range(1, 4):
                    map = [2,   5,  8,  1,  4,  7, 0,   3,  6,
                           11, 14, 17, 10, 13, 16, 9,  12, 15,
                           20, 23, 26, 19, 22, 25, 18, 21, 24]
                    for i in range(0, 27):
                        occupied1[i] = occupied[map[i]]
                    occupied = occupied1[:]
                    models.append({
                        "slot": slotid,
                        "block": blockid,
                        "occupied": "".join(occupied),
                        "rotation": (0, 0, math.pi*(-a)/2),
                        "random": block.vox_settings.random_count,
                        "weight": weight,
                        "ground": block.vox_settings.only_on_ground,
                        "even_x": block.vox_settings.even_x,
                        "even_y": block.vox_settings.even_y,
                        "even_z": block.vox_settings.even_z,
                    })

    return models


def rebuild_model(context):
    """Recalculate voxel models for all verticies"""
    obj = context.object
    origmode = obj.mode
    bpy.ops.object.mode_set(mode='EDIT')

    me = obj.data
    bm = bmesh.from_edit_mesh(me)

    kd = build_KDTree(bm)
    bm.verts.ensure_lookup_table()
    models = getmodels(context)

    if len(models) > 0:
        for i, v in enumerate(bm.verts):
            update_vertex_model(bm, kd, v, models, obj.vox_settings.zero_level_ground)
        bmesh.update_edit_mesh(obj.data)
    bpy.ops.object.mode_set(mode=origmode)


def rebuild_blocks(context, co1, co2):
    """Recalculate voxel models for specific verticies"""
    obj = context.object
    origmode = obj.mode
    bpy.ops.object.mode_set(mode='EDIT')

    me = obj.data
    bm = bmesh.from_edit_mesh(me)

    kd = build_KDTree(bm)
    bm.verts.ensure_lookup_table()
    models = getmodels(context)

    if len(models) > 0:
        """for x in range(round(co1[0]), round(co2[0])):
            for y in range(round(co1[1]), round(co2[1])):
                for z in range(round(co1[2]), round(co2[2])):
                    co, index, dist = kd.find(Vector((x, y, z)))
                    update_vertex_model(bm, kd, bm.verts[index], models, obj.vox_settings.zero_level_ground)
        """
        center = (Vector(co2) + Vector(co1))/2
        radius = (Vector(co2) - Vector(co1)).length/2 + 1
        vertexes = kd.find_range(center, radius)
        for co, index, dist in vertexes:
            update_vertex_model(bm, kd, bm.verts[index], models, obj.vox_settings.zero_level_ground)
    bpy.ops.object.mode_set(mode=origmode)


def cashe_vertexes(bm):
    out = {}
    for v in bm.verts:
        key = str(round(v.co[0])) + '-' + str(round(v.co[1])) + '-' + str(round(v.co[2]))
        out[key] = v
    return out


def build_KDTree(bm):
    """Builds KD Tree from bmesh"""
    size = len(bm.verts)
    kd = KDTree(size)
    for i, xyz in enumerate(bm.verts):
        kd.insert(xyz.co[:], i)
    kd.balance()

    return kd


def add_vertex_bm(bm, coord, color, object):
    """Adds vertex to bmesh at given coordianates with specific color"""
    vertex = bm.verts.new(coord)

    vg = bm.verts.layers.float_color['Vox Color']
    vm = bm.verts.layers.int['Vox Model']
    vertex[vg] = color[:]+(1,)
    vertex[vm] = object.vox_settings.active_slot

    return vertex


def change_slot(context, fr, to):
    """Change vox model slot for each vertex from 'fr' to 'to'"""
    obj = context.object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)
    vm = bm.verts.layers.int['Vox Model']

    for v in bm.verts:
        if v[vm] == fr:
            v[vm] = to

    bmesh.update_edit_mesh(obj.data)


def remove_vertex_volume(context, start, end):
    obj = context.object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)

    kd = build_KDTree(bm)

    x1, y1, z1 = round(min(start[0], end[0])), round(min(
        start[1], end[1])),  round(min(start[2], end[2]))
    x2, y2, z2 = round(max(start[0], end[0])), round(max(
        start[1], end[1])),  round(max(start[2], end[2]))

    rmlist = []
    for x in range(0, 2 if obj.use_mesh_mirror_x else 1):
        for y in range(0, 2 if obj.use_mesh_mirror_y else 1):
            for z in range(0, 2 if obj.use_mesh_mirror_z else 1):
                xs1 = x1 if x == 0 else -x2
                xs2 = x2 if x == 0 else -x1
                ys1 = y1 if y == 0 else -y2
                ys2 = y2 if y == 0 else -y1
                zs1 = z1 if z == 0 else -z2
                zs2 = z2 if z == 0 else -z1
                rmlist = remove_vertex_volume_fill(
                    xs1, xs2, ys1, ys2, zs1, zs2, kd, rmlist)

    rmlist.sort(reverse=True)

    for index in rmlist:
        bm.verts.ensure_lookup_table()
        bm.verts.remove(bm.verts[index])

    for x in range(0, 2 if obj.use_mesh_mirror_x else 1):
        for y in range(0, 2 if obj.use_mesh_mirror_y else 1):
            for z in range(0, 2 if obj.use_mesh_mirror_z else 1):
                xs1 = x1 if x == 0 else -x2
                xs2 = x2 if x == 0 else -x1
                ys1 = y1 if y == 0 else -y2
                ys2 = y2 if y == 0 else -y1
                zs1 = z1 if z == 0 else -z2
                zs2 = z2 if z == 0 else -z1
                rebuild_blocks(context, (xs1-1, ys1-1, zs1-1),  (xs2+2, ys2+2, zs2+2))

    bmesh.update_edit_mesh(obj.data)


def remove_vertex_volume_fill(x1, x2, y1, y2, z1, z2, kd, rmlist):
    for x in range(x1, x2+1):
        for y in range(y1, y2+1):
            for z in range(z1, z2+1):
                co, index, dist = kd.find(Vector((x, y, z)))
                if co:
                    if dist <= 0.5:
                        if index not in rmlist:
                            rmlist.append(index)

    return rmlist


def add_vertex_volume_fill(context, x1, x2, y1, y2, z1, z2, kd, bm, obj, color, fills):

    for x in range(x1, x2+1):
        for y in range(y1, y2+1):
            for z in range(z1, z2+1):
                if (x, y, z) not in fills:
                    fills.append((x, y, z))
                    co, index, dist = kd.find(Vector((x, y, z)))
                    if co:
                        if dist <= 0.5:
                            continue
                    add_vertex_bm(bm, Vector((x, y, z)), color, obj)

    rebuild_blocks(context, (x1-1, y1-1, z1-1),  (x2+2, y2+2, z2+2))

    return fills


def add_vertex_volume(context, start, end, color):
    """
    Adds vertex volume at spectific coordinates
    kd tree algorithm is used
    """
    obj = context.object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)
    vox_gn_setup.validate_attributes(bm)

    kd = build_KDTree(bm)

    x1, y1, z1 = round(min(start[0], end[0])), round(min(
        start[1], end[1])),  round(min(start[2], end[2]))
    x2, y2, z2 = round(max(start[0], end[0])), round(max(
        start[1], end[1])),  round(max(start[2], end[2]))
    fills = []

    for x in range(0, 2 if obj.use_mesh_mirror_x else 1):
        for y in range(0, 2 if obj.use_mesh_mirror_y else 1):
            for z in range(0, 2 if obj.use_mesh_mirror_z else 1):
                xs1 = x1 if x == 0 else -x2
                xs2 = x2 if x == 0 else -x1
                ys1 = y1 if y == 0 else -y2
                ys2 = y2 if y == 0 else -y1
                zs1 = z1 if z == 0 else -z2
                zs2 = z2 if z == 0 else -z1
                fills = add_vertex_volume_fill(
                    context, xs1, xs2, ys1, ys2, zs1, zs2, kd, bm, obj, color, fills)

    bmesh.update_edit_mesh(obj.data)


def get_height(context, start, stop, vector):
    """
    Calculates z-offset for extruding
    """
    region = context.region
    r3d = context.space_data.region_3d
    offset1 = stop-start
    object_matrix_inverted = copy.copy(context.object.matrix_world)

    layer0 = location_3d_to_region_2d(
        region, r3d, object_matrix_inverted @ Vector((0, 0, 0)))
    layer1 = location_3d_to_region_2d(
        region, r3d, object_matrix_inverted @ vector)
    offset2 = layer1 - layer0

    steps = round(offset1.dot(offset2) / offset2.length / offset2.length)

    return steps


def add_vertex(context, coord, color):
    """
    Adds vertex at spectific coordinate
    """
    add_vertex_volume(context, coord, coord, color)


def square_distance(vect1, vect2):
    return max(
        abs(vect1[0]-vect2[0]),
        abs(vect1[1]-vect2[1]),
        abs(vect1[2]-vect2[2])
    )


global kdcahsed


def get_color_prep(context):
    obj = context.object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)

    vox_gn_setup.validate_attributes(bm)
    kd = build_KDTree(bm)
    global kdcahsed
    kdcahsed = kd


def get_color(context, coord):
    """
    Get color from vertex
    :return: none
    """
    obj = context.object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)

    global kdcahsed
    kd = kdcahsed

    vg = bm.verts.layers.float_color['Vox Color']

    co, index, dist = kd.find(coord)
    bm.verts.ensure_lookup_table()
    return bm.verts[index][vg]


def set_color_prep(context):
    obj = context.object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)

    vox_gn_setup.validate_attributes(bm)
    kd = build_KDTree(bm)
    global kdcahsed
    kdcahsed = kd


def set_color_inradius(context, coord, color, sterngth, radius, falloff):
    """
    Set model to within the radius
    :return: none
    """
    obj = context.object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)

    global kdcahsed
    kd = kdcahsed

    vg = bm.verts.layers.float_color['Vox Color']
    for x in range(0, 2 if obj.use_mesh_mirror_x else 1):
        for y in range(0, 2 if obj.use_mesh_mirror_y else 1):
            for z in range(0, 2 if obj.use_mesh_mirror_z else 1):
                x1 = coord[0] if x == 0 else -coord[0]
                y1 = coord[1] if y == 0 else -coord[1]
                z1 = coord[2] if z == 0 else -coord[2]
                list = kd.find_range((x1, y1, z1), radius)
                for v in list:
                    edge = 1 - v[2]/radius
                    if edge < falloff:
                        edge = edge / falloff
                    else:
                        edge = 1
                    bm.verts.ensure_lookup_table()
                    bm.verts[v[1]][vg] = bm.verts[v[1]][vg].lerp(
                        color[:]+(1,), sterngth * edge)

    bmesh.update_edit_mesh(obj.data)


def set_model_inradius(context, coord, radius):
    """
    Set color to within the radius
    :return: none
    """
    obj = context.object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)

    global kdcahsed
    kd = kdcahsed

    vm = bm.verts.layers.int['Vox Model']
    for x in range(0, 2 if obj.use_mesh_mirror_x else 1):
        for y in range(0, 2 if obj.use_mesh_mirror_y else 1):
            for z in range(0, 2 if obj.use_mesh_mirror_z else 1):
                x1 = round(coord[0]) if x == 0 else round(-coord[0])
                y1 = round(coord[1]) if y == 0 else round(-coord[1])
                z1 = round(coord[2]) if z == 0 else round(-coord[2])
                list = kd.find_range((x1, y1, z1), radius)
                for v in list:

                    bm.verts.ensure_lookup_table()
                    bm.verts[v[1]][vm] = obj.vox_settings.active_slot

                rebuild_blocks(context, (x1-radius, y1-radius, z1-radius), (x1+radius, y1+radius, z1+radius))

    bmesh.update_edit_mesh(obj.data)


def set_model(context, coord):
    """
    Set model to vertex
    :return: none
    """
    obj = context.object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)

    global kdcahsed
    kd = kdcahsed

    vm = bm.verts.layers.int['Vox Model']

    for x in range(0, 2 if obj.use_mesh_mirror_x else 1):
        for y in range(0, 2 if obj.use_mesh_mirror_y else 1):
            for z in range(0, 2 if obj.use_mesh_mirror_z else 1):
                x1 = round(coord[0]) if x == 0 else round(-coord[0])
                y1 = round(coord[1]) if y == 0 else round(-coord[1])
                z1 = round(coord[2]) if z == 0 else round(-coord[2])
                co, index, dist = kd.find((x1, y1, z1))

                bm.verts.ensure_lookup_table()
                bm.verts[index][vm] = obj.vox_settings.active_slot

                rebuild_blocks(context, (x1-1, y1-1, z1-1), (x1+2, y1+2, z1+2))

    bmesh.update_edit_mesh(obj.data)


def set_color(context, coord, color, sterngth):
    """
    Set color to vertex
    :return: none
    """
    obj = context.object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)

    global kdcahsed
    kd = kdcahsed

    vg = bm.verts.layers.float_color['Vox Color']

    for x in range(0, 2 if obj.use_mesh_mirror_x else 1):
        for y in range(0, 2 if obj.use_mesh_mirror_y else 1):
            for z in range(0, 2 if obj.use_mesh_mirror_z else 1):
                x1 = coord[0] if x == 0 else -coord[0]
                y1 = coord[1] if y == 0 else -coord[1]
                z1 = coord[2] if z == 0 else -coord[2]
                co, index, dist = kd.find((x1, y1, z1))
                bm.verts.ensure_lookup_table()
                bm.verts[index][vg] = bm.verts[index][vg].lerp(
                    color[:]+(1,), sterngth)

    bmesh.update_edit_mesh(obj.data)


global cahsed_bvh_tree
cahsed_bvh_tree = None


def clearcashe():
    global cahsed_bvh_tree
    cahsed_bvh_tree = None


def build_bvh_tree(context):

    global cahsed_bvh_tree

    if cahsed_bvh_tree is None:
        obj = context.object
        me = obj.data
        bm = bmesh.from_edit_mesh(me)
        m = context.object.matrix_world
        vertices = []
        polygons = []

        for i, v in enumerate(bm.verts):

            co = Vector((round(v.co[0]), round(v.co[1]), round(v.co[2])))

            vertices.append(m @ co+Vector((0.5,  0.5,  0.5)))  # 0
            vertices.append(m @ co+Vector((0.5, -0.5,  0.5)))  # 1
            vertices.append(m @ co+Vector((-0.5, -0.5,  0.5)))  # 2
            vertices.append(m @ co+Vector((-0.5,  0.5,  0.5)))  # 3
            vertices.append(m @ co+Vector((0.5,  0.5, -0.5)))  # 4
            vertices.append(m @ co+Vector((0.5, -0.5, -0.5)))  # 5
            vertices.append(m @ co+Vector((-0.5, -0.5, -0.5)))  # 6
            vertices.append(m @ co+Vector((-0.5,  0.5, -0.5)))  # 7

            polygons.append((i*8+3, i*8+2, i*8+1, i*8+0))
            polygons.append((i*8+4, i*8+5, i*8+6, i*8+7))
            polygons.append((i*8+0, i*8+4, i*8+7, i*8+3))
            polygons.append((i*8+1, i*8+5, i*8+6, i*8+2))
            polygons.append((i*8+3, i*8+2, i*8+6, i*8+7))
            polygons.append((i*8+0, i*8+1, i*8+5, i*8+4))

        cahsed_bvh_tree = BVHTree.FromPolygons(vertices, polygons, all_triangles=False, epsilon=0)

    return cahsed_bvh_tree


def get_addvox_position(context, x, y, newp, exactLocation=False):
    """
    Method returns postion of new voxel that's going to be created
    at screen space coordinates - BVH tree used
    :return: vector3 position
    """
    region = context.region
    r3d = context.space_data.region_3d
    scene = context.scene

    view = region_2d_to_vector_3d(region, r3d, (x, y))
    loc = region_2d_to_location_3d(region, r3d, (x, y), view)
    ray_target = loc + (view * -100)
    ray_goal = loc + (view * 100000)

    bvhTree = build_bvh_tree(context)

    location, normal, index, distance = bvhTree.ray_cast(ray_target, ray_goal)

    # print(location, normal, index, distance)

    object_matrix_inverted = copy.copy(context.object.matrix_world)
    object_matrix_inverted.invert()

    if location is not None:
        location = object_matrix_inverted @ location

        xoff = getoffsetof1(location[0])
        yoff = getoffsetof1(location[1])
        zoff = getoffsetof1(location[2])

        delta = 0.001 if newp else 0
        vect = False
        if xoff + delta < yoff and xoff + delta < zoff:
            vect = Vector((1, 0, 0))
        elif yoff + delta < xoff and yoff + delta < zoff:
            vect = Vector((0, 1, 0))
        elif zoff + delta < xoff and zoff + delta < yoff:
            vect = Vector((0, 0, 1))

        if not vect:
            return False, False

        view = object_matrix_inverted.to_quaternion() @ view
        direction = dot(view, vect)
        # print(vect, view, vect, direction)

        if (direction > 0) != (newp is False):
            vect = -1*vect

        if not exactLocation:
            location = location + vect*0.5
            location = Vector(
                (round(location[0]), round(location[1]), round(location[2])))

        return location, vect

    else:
        ray_target = object_matrix_inverted @ ray_target
        ray_goal = object_matrix_inverted @ ray_goal
        location = world_plane_intersection(ray_target, ray_goal)
        location = Vector(
            (round(location[0]), round(location[1]), round(location[2])))

        return location, False


def get_addvox_position_b(context, x, y, newp, exactLocation=False):
    """
    Method returns postion of new voxel that's going to be created
    at screen space coordinates
    :return: vector3 position
    """
    region = context.region
    r3d = context.space_data.region_3d
    scene = context.scene

    view = region_2d_to_vector_3d(region, r3d, (x, y))
    loc = region_2d_to_location_3d(region, r3d, (x, y), view)
    ray_target = loc + (view * -100)
    ray_goal = loc + (view * 100000)
    depsgraph = context.view_layer.depsgraph

    result, location, normal, index, object, matrix = scene.ray_cast(
        depsgraph, ray_target, ray_goal)

    object_matrix_inverted = copy.copy(context.object.matrix_world)
    object_matrix_inverted.invert()

    if result:
        location = object_matrix_inverted @ location

        xoff = getoffsetof1(location[0])
        yoff = getoffsetof1(location[1])
        zoff = getoffsetof1(location[2])

        delta = 0.1 if newp else 0
        vect = False
        if xoff + delta < yoff and xoff + delta < zoff:
            vect = Vector((1, 0, 0))
        elif yoff + delta < xoff and yoff + delta < zoff:
            vect = Vector((0, 1, 0))
        elif zoff + delta < xoff and zoff + delta < yoff:
            vect = Vector((0, 0, 1))

        if not vect:
            return False, False

        view = object_matrix_inverted.to_quaternion() @ view
        direction = dot(view, vect)
        # print(vect, view, vect, direction)

        if (direction > 0) != (newp is False):
            vect = -1*vect

        if not exactLocation:
            location = location + vect*0.5
            location = Vector(
                (round(location[0]), round(location[1]), round(location[2])))

        return location, vect

    else:
        ray_target = object_matrix_inverted @ ray_target
        ray_goal = object_matrix_inverted @ ray_goal
        location = world_plane_intersection(ray_target, ray_goal)
        location = Vector(
            (round(location[0]), round(location[1]), round(location[2])))

        return location, False


def getoffsetof1(x):
    return abs(abs(x % 1)-0.5)


def dot(v1, v2):
    if len(v1) != len(v2):
        raise ValueError
    return sum(i1*i2 for i1, i2 in zip(v1, v2))


def get_plane_intersection(context, x, y, offset, exactLocation=False,
                           cursoroffset=Vector((0, 0, 0)),
                           drawnormal=Vector((0, 0, 1))):
    region = context.region
    r3d = context.space_data.region_3d

    view = region_2d_to_vector_3d(region, r3d, (x, y))
    loc = region_2d_to_location_3d(region, r3d, (x, y), view)
    ray_target = loc + (view * -10)
    ray_goal = loc + (view * 10000)

    object_matrix_inverted = copy.copy(context.object.matrix_world)
    object_matrix_inverted.invert()
    ray_target = object_matrix_inverted @ ray_target
    ray_goal = object_matrix_inverted @ ray_goal
    if abs(drawnormal[0]) > 0.9:
        # yz - plane
        location = world_plane_intersection(
            ray_target, ray_goal, offset, cursoroffset, "YZ")
    elif abs(drawnormal[1]) > 0.9:
        # xz - plane
        location = world_plane_intersection(
            ray_target, ray_goal, offset, cursoroffset, "XZ")
    else:
        # xy - plane
        location = world_plane_intersection(
            ray_target, ray_goal, offset, cursoroffset, "XY")
    if not exactLocation:
        location = Vector(
            (round(location[0]), round(location[1]), round(location[2])))

    return location


def world_plane_intersection(orign, target, offset=Vector((0, 0, 0)),
                             cursoroffset=Vector((0, 0, 0)), plane="XY"):
    orign = orign - offset - cursoroffset
    target = target - offset - cursoroffset
    x1, y1, z1 = orign[0], orign[1], orign[2]
    x2, y2, z2 = target[0], target[1], target[2]

    # seems to be a side view
    if abs(z1 - z2) > abs(1000*(x1 - x2)) + abs(1000*(y1 - y2)):
        return Vector((x1, y1, 0)) + offset
    if abs(x1 - x2) > abs(1000*(z1 - z2)) + abs(1000*(y1 - y2)):
        return Vector((0, y1, z1)) + offset
    if abs(y1 - y2) > abs(1000*(x1 - x2)) + abs(1000*(z1 - z2)):
        return Vector((x1, 0, z1)) + offset

    else:
        if plane == "YZ":
            z = -x1 * (z2-z1) / (x2-x1) + z1
            y = -x1 * (y2-y1) / (x2-x1) + y1
            return Vector((0, y, z)) + offset
        elif plane == "XZ":
            x = -y1 * (x2-x1) / (y2-y1) + x1
            z = -y1 * (z2-z1) / (y2-y1) + z1
            return Vector((x, 0, z)) + offset
        else:
            x = -z1 * (x2-x1) / (z2-z1) + x1
            y = -z1 * (y2-y1) / (z2-z1) + y1
            return Vector((x, y, 0)) + offset
