from mathutils import Vector
import gpu
import bgl
import copy
import bpy
import mathutils
import math
from gpu_extras.batch import batch_for_shader
from bpy_extras.view3d_utils import location_3d_to_region_2d

from . import vox_manager

global draw_handler
draw_handler = 0

global last_pos
last_pos = Vector((10000, 0, 0))

global last_vect
last_vect = False


def resetHelper():
    """Removes voxel preview"""
    global draw_handler
    global last_pos
    global last_vect

    last_pos = Vector((10000, 0, 0))
    last_vect = False

    if(draw_handler != 0):
        bpy.types.SpaceView3D.draw_handler_remove(
            draw_handler, 'WINDOW')
    draw_handler = 0


def drawCircle(location, context, radius):
    """Adds voxel preview in viewport"""

    resetHelper()
    global draw_handler
    draw_handler = bpy.types.SpaceView3D.draw_handler_add(
        draw_circle,
        (location, context, radius),
        'WINDOW', 'POST_PIXEL'
    )
    context.area.tag_redraw()


def draw_circle(location, context, radius):
    if bpy.context.mode != 'EDIT_MESH':
        return

    vertices = []
    for i in range(0, 50):
        vertices.append(
            Vector((
                math.sin(i/25*math.pi) * radius + location[0],
                math.cos(i/25*math.pi) * radius + location[1]
            )))

    shader = gpu.shader.from_builtin('2D_UNIFORM_COLOR')
    batch = batch_for_shader(
        shader, 'LINES', {"pos": vertices})

    shader.bind()
    shader.uniform_float("color", (1, 1, 1, 1))
    batch.draw(shader)


def draw_occupied_preview(context):
    region = context.region
    r3d = context.space_data.region_3d
    radius = 5
    res = 24

    obj = context.object
    if obj is None:
        return
    vs = obj.vox_settings

    object_matrix = copy.copy(context.object.matrix_world)

    for x in range(-1, 2):
        for y in range(-1, 2):
            for z in range(-1, 2):
                if x == 0 and y == 0 and z == 0:
                    continue

                id = vox_manager.vox_offset_to_id(x, y, z)
                state = getattr(vs, "occupied"+str(id))

                if(state == "DEF"):
                    continue

                v = Vector((x, y, z))
                v = object_matrix @ v
                location = location_3d_to_region_2d(region, r3d, v)
                if location is None:
                    return
                color = (1, 1, 1, 0.5)
                vertices = []
                bgl.glLineWidth(3)
                mode = 'LINES'

                if state == "EMPTY":
                    bgl.glLineWidth(4)
                    vertices.append(Vector((radius + location[0], location[1])))
                    vertices.append(Vector((-radius + location[0],  location[1])))

                if state == "FREE":
                    vertices.append(Vector((radius + location[0], radius + location[1])))
                    vertices.append(Vector((-radius + location[0], -radius + location[1])))
                    vertices.append(Vector((-radius + location[0], radius + location[1])))
                    vertices.append(Vector((radius + location[0], -radius + location[1])))

                if state == "SAME":
                    bgl.glLineWidth(2)
                    mode = 'TRIS'
                    indices = set()
                    vertices.append(Vector((location[0], location[1])))
                    for i in range(0, res):

                        vertices.append(
                            Vector((
                                math.cos(i/res*2*math.pi) * radius + location[0],
                                math.sin(i/res*2*math.pi) * radius + location[1]
                            )))

                        if(i < res-1):
                            indices.add((0, i+1, i+2))
                        else:
                            indices.add((0, i+1, 1))
                    color = (1, 0, 0, 0.5)

                if state == "ANY":
                    for i in range(0, res):
                        vertices.append(
                            Vector((
                                math.sin(i/res*2*math.pi) * radius + location[0],
                                math.cos(i/res*2*math.pi) * radius + location[1]
                            )))
                        vertices.append(
                            Vector((
                                math.sin((i-1)/res*2*math.pi) * radius + location[0],
                                math.cos((i-1)/res*2*math.pi) * radius + location[1]
                            )))

                    color = (1, 0, 0, 0.5)

                shader = gpu.shader.from_builtin('2D_UNIFORM_COLOR')

                if mode == 'TRIS':
                    batch = batch_for_shader(
                        shader, mode, {"pos": vertices}, indices=indices)
                else:
                    batch = batch_for_shader(
                        shader, mode, {"pos": vertices})

                shader.bind()
                shader.uniform_float("color", color)
                batch.draw(shader)


def occupied_preview(context):
    """Adds helper to understand occupied settings"""

    resetHelper()
    global draw_handler

    draw_handler = bpy.types.SpaceView3D.draw_handler_add(
        draw_occupied_preview,
        (context,),
        'WINDOW', 'POST_PIXEL'
    )


def setHelper(location, context, shape, vect, color):
    """Adds voxel preview in viewport"""

    global last_pos
    global last_vect

    if(location == last_pos) and (vect == last_vect):
        return False

    last_pos = location
    last_vect = vect

    resetHelper()
    global draw_handler
    draw_handler = bpy.types.SpaceView3D.draw_handler_add(
        draw_helper,
        (location, context, shape, vect, color),
        'WINDOW', 'POST_VIEW'
    )


def draw_helper(loc, context, model, normal, color):
    """Draw helper in viewport"""

    if bpy.context.mode != 'EDIT_MESH':
        return

    if(model == "PLANE"):
        vertices = gen_plane_coords(loc, normal)
    elif(model == "VOLUME"):
        vertices = gen_volume_coords(loc, normal)
    else:
        vertices = gen_volume_coords(loc, loc)

    object_matrix = copy.copy(context.object.matrix_world)

    for i, v in enumerate(vertices):
        vertices[i] = object_matrix @ v
    bgl.glEnable(bgl.GL_BLEND)
    shader = gpu.shader.from_builtin('3D_UNIFORM_COLOR')

    bgl.glLineWidth(3)
    batch = batch_for_shader(
        shader, 'LINES', {"pos": vertices})
    shader.bind()
    shader.uniform_float("color", (color[0], color[1], color[2], 0.8))

    batch.draw(shader)

    context.area.tag_redraw()


def gen_volume_coords(loc1, loc2):
    """Point coordinates for box given size"""
    x1, y1, z1 = min(loc1[0], loc2[0])-0.5, min(
        loc1[1], loc2[1])-0.5, min(loc1[2], loc2[2])-0.5
    x2, y2, z2 = max(loc1[0], loc2[0])+0.5, max(
        loc1[1], loc2[1])+0.5, max(loc1[2], loc2[2])+0.5

    vertices = [
        Vector((x2, y2, z2)),  Vector((x2, y2, z1)),
        Vector((x2, y2, z2)),  Vector((x2, y1, z2)),
        Vector((x2, y2, z2)),  Vector((x1, y2, z2)),
        Vector((x1, y1, z1)),  Vector((x1, y2, z1)),
        Vector((x1, y1, z1)),  Vector((x2, y1, z1)),
        Vector((x1, y1, z1)),  Vector((x1, y1, z2)),
        Vector((x2, y1, z1)),  Vector((x2, y1, z2)),
        Vector((x2, y1, z1)),  Vector((x2, y2, z1)),
        Vector((x1, y2, z1)),  Vector((x1, y2, z2)),
        Vector((x1, y2, z1)),  Vector((x2, y2, z1)),
        Vector((x1, y1, z2)),  Vector((x1, y2, z2)),
        Vector((x1, y1, z2)),  Vector((x2, y1, z2)),
    ]
    return vertices


def gen_plane_coords(loc, normal):
    """Plane coordinates in location rotated by normal"""
    if abs(normal[0]) > abs(normal[1]) and abs(normal[0]) > abs(normal[2]):
        vertices = [
            loc + Vector((0, 0.5, 0.5)), loc + Vector((0, 0.5, -0.5)),
            loc + Vector((0, 0.5, 0.5)), loc + Vector((0, -0.5, 0.5)),
            loc + Vector((0, -0.5, -0.5)), loc + Vector((0, 0.5, -0.5)),
            loc + Vector((0, -0.5, -0.5)), loc + Vector((0, -0.5, 0.5)),
        ]
    elif abs(normal[1]) > abs(normal[0]) and abs(normal[1]) > abs(normal[2]):
        vertices = [
            loc + Vector((0.5, 0, 0.5)), loc + Vector((0.5, 0, -0.5)),
            loc + Vector((0.5, 0, 0.5)), loc + Vector((-0.5, 0, 0.5)),
            loc + Vector((-0.5, 0, -0.5)), loc + Vector((0.5, 0, -0.5)),
            loc + Vector((-0.5, 0, -0.5)), loc + Vector((-0.5, 0, 0.5)),
        ]
    else:
        vertices = [
            loc + Vector((0.5, 0.5, 0)), loc + Vector((0.5, -0.5, 0)),
            loc + Vector((0.5, 0.5, 0)), loc + Vector((-0.5, 0.5, 0)),
            loc + Vector((-0.5, -0.5, 0)), loc + Vector((0.5, -0.5, 0)),
            loc + Vector((-0.5, -0.5, 0)), loc + Vector((-0.5, 0.5, 0)),
        ]
    for i, v in enumerate(vertices):
        vertices[i] = v - normal/2

    return vertices
